# =============================================================================
# @file    __version__.py
# @brief   stochss-compute version info
# @license Please see the file named LICENSE.md in parent directory
# @website https://github.com/StochSS/stochss-compute
# =============================================================================


__version__      = "0.9.5"

__title__        = "stochss-compute"
__description__  = "A compute delegation package for the StochSS family of stochastic simulators"
__url__          = "https://github.com/StochSS/stochss-compute"
__download_url__ = "https://pypi.org/project/stochss-compute/#files"
__author__       = "See CONTRIBUTORS.md"
__email__        = "briandrawert@gmail.com"
__license__      = "GPLv3"
__copyright__    = "Copyright (C) 2017-2022"
